//
//  Project   : besy
//  File      : TRSA.hh
//  Author    : Ronald Kriemann
//  Purpose   : class for RSA encoding and decoding
//  Date      : 01.12.1997
//

extern "C"
{
#include <unistd.h>
}

#include "defines.hh"
#include "base64.hh"
#include "md5.hh"
#include "timer.hh"
#include "TRandom.hh"
#include "TBitArray.hh"
#include "idea.hh"

#include "TRSA.hh"

#define DELIM '#'

// begin of public/private sign
#define BEGIN_SIGN  "--SIGN--"

//
// constructor
//
TRSA::TRSA ()
{
    // defined public key part
    e = 3;

    verbosity = 0;
}

//
// destructor
//
TRSA::~TRSA ()
{
}

//
// access keys
//
TBigInteger<RSA_BASE> & 
TRSA::getPrivateKey ()
{
    return d;
}

TBigInteger<RSA_BASE> & 
TRSA::getPublicKey ()
{
    return e;
}

TBigInteger<RSA_BASE> & 
TRSA::getModulus ()
{
    return n;
}

void 
TRSA::setPrivateKey ( const TBigInteger<RSA_BASE> &k )
{
    d = k;
}

void 
TRSA::setPublicKey ( const TBigInteger<RSA_BASE> &k )
{
    e = k;
}

void 
TRSA::setModulus ( const TBigInteger<RSA_BASE> &k )
{
    n = k;
}

//
// build keys
//
void 
TRSA::buildKeys ()
{
    uint    j, maxSteps;
    bool    bFinished, bHaveKeys = FALSE;
        
    TBigInteger<RSA_BASE>   max, temp, t;

    // our two primes
    TBigInteger<RSA_BASE>   p, q;
    TBitArray               bitArr;

    // defined public key part
    e = 3;

    p = q = 2;

    maxSteps = 500;
    
    //
    // build a first prime
    //        
    bFinished = FALSE;
    j = maxSteps+1;
    while (!bFinished)
    {
        if (p.isPrime())
        {
            bFinished = TRUE;
            if (verbosity > 0)
                cout << "|" << endl;
        }// if
        else
        {
            j++;
            if (j > maxSteps)
            {
                if (verbosity > 0)
                    cout << "*" << flush;

                // init p randomly
                bitArr.setSize( KEY_SIZE / 2 );
                bitArr.initRandom();
                bitArr.setBit( 0, 1 );
                bitArr.setBit( (KEY_SIZE / 2)-1, 1 );
                bitArr.setBit( (KEY_SIZE / 2)-2, 1 );

                p.fromString( bitArr.toString() );

                j = 0;
            }// if
            else
            {
                p += 2;
                if (verbosity > 0)
                    cout << "-" << flush;
            }// else
        }// else
    }// while
    
    while (!bHaveKeys)
    {
        //
        // build second prime
        //
        
        bFinished = FALSE;
        j  = maxSteps + 1;
        q += 2;
        while (!bFinished)
        {
            if (q.isPrime())
            {
                bFinished = TRUE;
                if (verbosity > 0)
                    cout << "|" << endl;
            }// if
            else
            {
                j++;
                if (j > maxSteps)
                {
                    if (verbosity > 0)
                        cout << "*" << flush;

                    bitArr.setSize( KEY_SIZE / 2 );
                    bitArr.initRandom();
                    bitArr.setBit( 0, 1 );
                    bitArr.setBit( (KEY_SIZE / 2)-1, 1 );
                    bitArr.setBit( (KEY_SIZE / 2)-2, 1 );
                    
                    q.fromString( bitArr.toString() );
                    
                    j = 0;
                }// 
                else
                {
                    q += 2;
                    if (verbosity > 0)
                        cout << "-" << flush;
                }// if
            }// else
        }// while
        
        // set n = pq
        n  = p;
        n *= q;
        
        //
        // determine d while building the mult. inv. in Z((p-1)*(q-1))
        //
        
        max  = p;
        max -= 1;
        t    = q;
        t   -= 1;
        max *= t;

        d = e.invMod( max );
        
        temp  = d;
        temp *= e;
        temp %= max;

        // check if mult. inv.
        if ((temp == 1) && (d != 1)) 
            bHaveKeys = TRUE;
        else
            p = q;  // just recalc q
    }// while
}

//
// encrypt a message (size < KEY_SIZE / TString::charSize)
//

TString
TRSA::encryptPrivate ( const TString &input ) const
{
    TString  buffer;
    TTimer   timer;

    if (verbosity > 0)
        cout << "TRSA::encrypt (...) : doing powerMod ";

    timer.start();
    buffer = encrypt( d, input );
    timer.stop();

    if (verbosity > 0)
        cout << "- done (in " << timer << ')' << endl;

    return buffer;
}

TString
TRSA::encryptPublic ( const TString &input ) const
{
    TString  buffer;

    buffer = encrypt( e, input );
    return buffer;
}

TString 
TRSA::encrypt ( const TBigInteger<RSA_BASE> &k,    const TString &input ) const
{
    TBigInteger<RSA_BASE>   msg;
    TBitArray               ideaKey;
    TString                 string, key;
    TIdea                   idea;

    msg.fromString( input );

    if (msg >= n)
    {
        ideaKey.setSize( KEY_SIZE - TString::charSize() );
        ideaKey.initRandom();

        key = ideaKey.toString();
        idea.generateSubKeys( key );
        
        key = encrypt( k, key );
        key.setLength( KEY_SIZE / TString::charSize() );

        string  = key;
        string += idea.encode( input );
    }// if
    else
    {
        msg.powerMod( k, n );

        string = msg.toString();
    }// else

    return string;
}

//
// decrypt a message (size < KEY_SIZE / 8)
//

TString
TRSA::decryptPrivate ( const TString &input ) const
{
    TString  buffer;
    TTimer   timer;

    if (verbosity > 0)
        cout << "TRSA::decrypt (...) : doing powerMod ";

    timer.start();
    buffer = decrypt( d, input );
    timer.stop();

    if (verbosity > 0)
        cout << "- done (in " << timer << ')' << endl;

    return buffer;
}

TString
TRSA::decryptPublic ( const TString &input ) const
{
    TString  buffer;

    buffer = decrypt( e, input );
    return buffer;
}

TString
TRSA::decrypt ( const TBigInteger<RSA_BASE> &k,    const TString &input ) const
{
    TBigInteger<RSA_BASE>   msg;
    TString                 string, key;
    TIdea                   idea;

    msg.fromString( input );

    if (msg >= n)
    {
        string = input;
        key = string.substr( 0, KEY_SIZE / TString::charSize() );
        string <<= KEY_SIZE / TString::charSize();

        key = decrypt( k, key );
        idea.generateSubKeys( key );

        string = idea.decode( string );
    }// if
    else
    {    
        msg.powerMod( k, n );

        string = msg.toString();
    }// else

    return string;
}

//
// sign a message (size < KEY_SIZE / 8)
//

TString
TRSA::sign ( const TString &input ) const
{
    TString  string;
    TMD5     md5;

    string  = input;
    string += '\n';
    string += BEGIN_SIGN;
    string += '\n';
    string += TBase64::encode( encryptPrivate( md5.crypt( input ) ) );

    return string;
}

//
// check a signed msg (assuming : sign at end)
//
bool
TRSA::checkSign ( const TString & signedMsg ) const
{
    TString  msg, subStr;
    TMD5     md5;

    msg = signedMsg;

    // extract sign from msg
    subStr = msg.strstr( BEGIN_SIGN );

    // remove sign from msg
    msg.setLength( msg.length() - subStr.length() - 1 );

    subStr.strtok( '\n' );
    
    if (subStr == "")
        return FALSE;

    if (subStr[subStr.length()-1] == '\n')
        subStr.setLength( subStr.length() - 1 );

    // check sign
    if (md5.crypt( msg ) == decryptPublic( TBase64::decode( subStr )))
        return TRUE;
    else
        return FALSE;
}

//
// remove sign from msg (assuming : sign at end)
//
void
TRSA::removeSign ( TString & msg ) const
{
    ulong    oldLength = msg.length();
    TString  string, subStr;

    string = msg;
    msg    = "";

    while (string.length() > 0)
    {
        subStr = string.strtok( '\n' );

        // check if string contains \n
        if (oldLength == string.length())
        {
            msg += string;
            return;
        }// if
        else
            oldLength = string.length();
        
        if (subStr == BEGIN_SIGN)
        {
            // remove \n at end
            msg.setLength( msg.length()-1 );

            return;
        }// if
        else
        {
            msg += subStr;
            msg += '\n';
        }// else
    }// while

    if (string != "")
        msg += string;
}
    
//
// return fingerprint of public key
//
TString
TRSA::fingerPrint () const
{
    TMD5     md5;

    return TBase64::encode( md5.crypt( n.toString() ) );
}
    
//
// set verbosity level
//
void
TRSA::setVerbosity ( int n )
{
    verbosity = n;
}

//
// output
//
ostream & 
operator << ( ostream &os, TRSA &rsa )
{
    os << "e = " << TBase64::encode( rsa.getPublicKey().toString() ) << endl;
    os << "d = " << TBase64::encode( rsa.getPrivateKey().toString() ) << endl;
    os << "n = " << TBase64::encode( rsa.getModulus().toString() ) << endl;
    os << "f = " << rsa.fingerPrint() << endl;
    
    return os;
}
